""" Test for the core code of the method module
"""
import numpy as np
import adafdr.method as md
import adafdr.data_loader as dl

def test_method_init():
    """ test for md.method_init
    """
    p, x, h, n_full, _ = dl.load_2d_bump_slope(n_sample=20000)
    a, mu, sigma, w = md.method_init(p, x, 2, alpha=0.1, n_full=n_full, h=h,
                                     random_state=0, fold_number=0)
    t = md.f_all(x, a, mu, sigma, w)
    gamma = md.rescale_mirror(t, p, 0.1)
    t = t*gamma
    n_rej = np.sum(p < t)
    FDP = np.sum((p < t)*(h == 0))/n_rej
    print('n_rej:', n_rej)
    assert n_rej > 700
    assert FDP < 0.15
    mu_ref1 = np.array([[0.25, 0.25], [0.75, 0.75]], dtype=float)
    mu_ref2 = np.array([[0.75, 0.75], [0.25, 0.25]], dtype=float)
    error = np.min([np.linalg.norm(mu-mu_ref1),
                    np.linalg.norm(mu-mu_ref2)])
    print('error for estimating mu = %0.8f'%error)
    assert error < 0.05
def test_reparametrize():
    """ test for md.reparametrize
    """
    w_init = np.array([0.4, 0.3, 0.3], dtype=float)
    a_init = np.array([2, 0.1], dtype=float)
    mu_init = np.array([[0.2, 0.2], [0.7, 0.7]], dtype=float)
    sigma_init = np.array([[0.1, 0.2], [0.1, 0.1]], dtype=float)
    d = 2
    x_test = np.array([[0.1, 0.2], [0.3, 0.5]], dtype=float)
    a, b, w, mu, sigma = md.reparametrize(a_init, mu_init, sigma_init, w_init, d)
    t_init = md.f_all(x_test, a_init, mu_init, sigma_init, w_init)
    t = md.t_cal(x_test, a, b, w, mu, sigma)
    print('t_init:', t_init)
    print('t', t)
    assert all(np.absolute(t_init-t) < 1e-8)
def test_rescale_mirror():
    """ test for md.rescale_mirror
    """
    p, x, _, _, _ = dl.load_2d_bump_slope(n_sample=2000)
    alpha = 0.1
    t = np.ones([x.shape[0]], dtype=float)
    gamma_grid = np.linspace(1e-4, 0.01, 100)
    alpha_hat = np.zeros([gamma_grid.shape[0]], dtype=float)
    for i in range(gamma_grid.shape[0]):
        alpha_hat[i] = np.sum(p > 1-t*gamma_grid[i])/np.sum(p < t*gamma_grid[i])
    gamma = np.max(gamma_grid[alpha_hat < alpha])
    gamma_test = md.rescale_mirror(t, p, alpha)
    print('gamma_GT', gamma)
    print('gamma_test', gamma_test)
    assert np.absolute(gamma-gamma_test) < 1e-4
def test_method_single_fold():
    """ test for md.method_single_fold
    """
    p, x, h, n_full, _ = dl.load_2d_bump_slope(n_sample=20000)
    n_rej, t, _ = md.method_single_fold(p, x, 2, alpha=0.1, n_full=n_full,
                                        n_itr=100, h=h, fold_number=0, random_state=0)
    FDP = np.sum((p < t)*(h == 0))/n_rej
    print('n_rej:', n_rej)
    assert n_rej > 800
    print('FDP:', n_rej)
    assert FDP < 0.15
def test_preprocess_two_fold():
    """ Test for preprocess_two_fold
    """
    np.random.seed(0)
    x_test_1 = np.random.choice([0, 1, 2, 3], size=300)
    x_test_2 = np.array([0, 1, 2, 3]).reshape([-1, 1])
    temp = np.arange(300)
    p_test = np.ones([300], dtype=float)
    p_test[x_test_1 == 0] = 0.001
    p_test[(x_test_1 == 1)*(temp < 200)] = 0.001
    p_test[(x_test_1 == 2)*(temp < 100)] = 0.001
    _, x_test_new_2 = md.preprocess_two_fold(p_test,
                                             x_test_1.reshape([-1, 1]),
                                             x_test_2,
                                             300, None)
    print('x_test_2', x_test_2)
    print('x_test_new_2', x_test_new_2)
    assert x_test_new_2[0] > 0.75
    assert (x_test_new_2[1] > 0.5) and (x_test_new_2[1] < 0.75)
    assert (x_test_new_2[2] > 0.25) and (x_test_new_2[2] < 0.5)
    assert x_test_new_2[3] < 0.25
def test_adafdr_test():
    """ Test for adafdr_test
    """
    p, x, h, n_full, _ = dl.load_2d_bump_slope(n_sample=20000)
    res = md.adafdr_test(p, x, K=2, alpha=0.1, h=None, n_full=n_full,\
                         n_itr=50, verbose=False, random_state=0,\
                         single_core=True)
    t = res['t_rej']
    FDP = np.sum((p < t)*(h == 0))/np.sum(p < t)
    n_rej = np.sum(p < t)
    print('n_rej', n_rej)
    assert n_rej > 700
    print('FDP', FDP)
    assert FDP < 0.12
    


import numpy as np
import scipy as sp
from scipy import stats
import matplotlib.pyplot as plt
from scipy.stats import rankdata
import logging
import pickle
import os

""" 
    basic functions
"""
def get_grid_1d(n_grid):    
    """
    return an equally spaced covariate covering the 1d space...

    Parameters
    ----------
    n_grid: int
        number of points 

    Returns
    -------
    (n,1) ndarray
    """
    x_grid = np.linspace(0,1,n_grid).reshape(-1,1)
    return x_grid

def get_grid_2d(n_grid):
    """
    return an equally spaced covariate covering the 2d space...

    Parameters
    ----------
    n_grid: int
        number of points 

    Returns
    -------
    (n,2) ndarray
    """
    temp = np.linspace(0,1,n_grid)
    g1,g2  = np.meshgrid(temp,temp)
    x_grid = np.concatenate([g1.reshape(-1,1),g2.reshape(-1,1)],axis=1)
    return x_grid

"""
    calculate the dimension-wise rank statistics
    # fix it: for discrete features, it may be nice to keep their values the same
    
    ----- input  -----
    x: an n*d array 
    
    ----- output -----
    ranks: an n*d array, column-wise rank of x
"""
def rank(x, continous_rank=True):
    """Calculate the dimension-wise rank statistics.
    
    Args:
        x ((n,d) ndarray): The covariates.
        continous_rank (bool): Indicate if break the same value by randomization.
    
    Returns:
        ranks ((n,d) ndarray): The column-wise rank of x
    """
    ranks = np.empty_like(x)
    n,d = x.shape
    for i in range(d):
        if continous_rank:           
            temp = x[:,i].argsort(axis=0)       
            ranks[temp,i] = np.arange(n)
        else:
            ranks[:,i] = rankdata(x[:,i])-1
    return ranks

def result_summary(pred, h=None, f_write=None, title=''):
    """ Summerize the result based on the predicted value and the true value
    
    Args:
        pred ((n,) ndarray): the testing result, 1 for alternative and 0 for null.
        h ((n,) ndarray): the true values.
        f_write (file handle)
        
    """
    if title != '':
        print('## %s'%title)
    print('# Num of discovery: %d'%np.sum(pred))
    if h is not None: 
        print("# Num of alternatives:",np.sum(h))
        print("# Num of true discovery: %d"%np.sum(pred*h))
        print("# Actual FDP: %0.3f"%(1-np.sum(pred * h) / np.sum(pred)))
    print('')    
    if f_write is not None:
        f_write.write('# Num of discovery: %d\n'%np.sum(pred))
        if h is not None:
            f_write.write("# Num of alternatives: %d\n"%np.sum(h))
            f_write.write("# Num of true discovery: %d\n"%np.sum(pred*h))
            f_write.write("# Actual FDP: %0.3f\n"%(1-np.sum(pred * h) / np.sum(pred)))
        f_write.write('\n')
    return

def print_param(a,mu,sigma,w):
    print('# w=%s'%w)
    print('# a=%s'%a)
    print('# mu=%s'%mu)
    print('# sigma=%s'%sigma)
    print('')

"""
    basic functions for visualization
""" 
def plot_x(x,vis_dim=None):
    if len(x.shape)==1:
        plt.hist(x,bins=50)
    else:
        if vis_dim is None: vis_dim = np.arange(x.shape[1])            
        for i,i_dim in enumerate(vis_dim):
            plt.subplot('1'+str(len(vis_dim))+str(i+1))
            plt.hist(x[:,i_dim],bins=50)
            plt.title('dimension %s'%str(i_dim+1))   
    
def plot_t(t,p,x,h=None,color=None,label=None):
    if color is None: color = 'darkorange'
        
    if t.shape[0]>5000:
        rand_idx=np.random.permutation(x.shape[0])[0:5000]
        t = t[rand_idx]
        p = p[rand_idx]
        x = x[rand_idx]
        if h is not None: h = h[rand_idx]
            
    if len(x.shape)==1:
        sort_idx = x.argsort()
        if h is None:
            plt.scatter(x,p,alpha=0.1,color='royalblue')
        else:
            plt.scatter(x[h==0],p[h==0],alpha=0.1,color='royalblue')
            plt.scatter(x[h==1],p[h==1],alpha=0.1,color='seagreen')
        plt.plot(x[sort_idx],t[sort_idx],color=color,label=label)
        plt.ylim([0,2*t.max()])
            
    else:
        n_plot = min(x.shape[1], 5)
        for i in range(n_plot):
            plt.subplot(str(n_plot)+'1'+str(i+1))
            sort_idx=x[:,i].argsort()
            if h is None:
                plt.scatter(x[:,i],p,alpha=0.1)
            else:
                plt.scatter(x[:,i][h==0],p[h==0],alpha=0.1,color='royalblue')
                plt.scatter(x[:,i][h==1],p[h==1],alpha=0.1,color='seagreen')
            plt.scatter(x[:,i][sort_idx],t[sort_idx],s=8,alpha=0.2,color='darkorange')
            plt.ylim([0,2*t.max()])
    #plt.scatter(x,t,alpha=0.2)
    #plt.ylim([0,1.2*t.max()])
    #plt.ylabel('t')
    #plt.xlabel('x')
    
def plot_scatter_t(t,p,x,h=None,color='orange',label=None):        
    if t.shape[0]>5000:
        rand_idx=np.random.permutation(x.shape[0])[0:5000]
        t = t[rand_idx]
        p = p[rand_idx]
        x = x[rand_idx]
        if h is not None: 
            h = h[rand_idx]            
    sort_idx = x.argsort()
    if h is None:
        plt.scatter(x,p,alpha=0.1,color='steelblue')
    else:
        plt.scatter(x[h==0],p[h==0],alpha=0.1,color='steelblue')
        plt.scatter(x[h==1],p[h==1],alpha=0.3,color='seagreen')
    plt.scatter(x[sort_idx],t[sort_idx],color=color,s=4,alpha=0.6,label=label)
    plt.ylim([0, 1.5*t.max()])
        
def plot_data_1d(p,x,h,n_pt=1000):
    rnd_idx=np.random.permutation(p.shape[0])[0:n_pt]
    p = p[rnd_idx]
    x = x[rnd_idx]
    h = h[rnd_idx]
    plt.scatter(x[h==1],p[h==1],color='r',alpha=0.2,label='alt')
    plt.scatter(x[h==0],p[h==0],color='b',alpha=0.2,label='null')
    plt.xlabel('covariate')
    plt.ylabel('p-value')
    plt.title('hypotheses') 
    
def plot_data_2d(p,x,h,n_pt=1000):
    rnd_idx=np.random.permutation(p.shape[0])[0:n_pt]
    p = p[rnd_idx]
    x = x[rnd_idx,:]
    h = h[rnd_idx]
    plt.scatter(x[h==1,0],x[h==1,1],color='r',alpha=0.2,label='alt')
    plt.scatter(x[h==0,0],x[h==0,1],color='b',alpha=0.2,label='null')
    plt.xlabel('covariate 1')
    plt.ylabel('covariate 2')
    plt.title('hypotheses') 

"""
    ancillary functions
""" 
def sigmoid(x):
    x = x.clip(min=-20,max=20)
    return 1/(1+np.exp(-x))
        

def inv_sigmoid(w):
    w = w.clip(min-1e-8,max=1-1e-8)
    return np.log(w/(1-w))

"""
    Functions for generating the simulation results
"""
def get_summary_stats(filename=None, folder_r=None):
    """Extract the statstics from the simulation results
    
    Args:
        filename (str): file path for the python results.
        folder_r (str): result for r methods.
    Return:
        summary_stats (dic): a dic containing FDP and Power.
    """
    summary_stats = {}
    # Python methods
    if filename is not None:
        fil = open(filename, 'rb')
        result_dic = pickle.load(fil)
        time_dic = pickle.load(fil)
        method_list = list(result_dic.keys())
        alpha_list = np.array([0.05, 0.1, 0.15, 0.2])
        n_data = len(result_dic[method_list[0]][alpha_list[0]])
        for method in method_list:
            summary_stats[method] = {}
            summary_stats[method]['FDP'] = np.zeros([n_data, len(alpha_list)])
            summary_stats[method]['Power'] = np.zeros([n_data, len(alpha_list)])
            for i_alpha,alpha in enumerate(alpha_list):
                for i_data,data in enumerate(result_dic[method][alpha]):
                    h, h_hat = data
                    summary_stats[method]['FDP'][i_data, i_alpha] =\
                        np.sum((h==0)*(h_hat==1)) / max(np.sum(h_hat==1), 1)
                    summary_stats[method]['Power'][i_data, i_alpha] =\
                        np.sum((h==1)*(h_hat==1)) / np.sum(h==1)
    # R methods
    if folder_r is not None:
        # file_list = os.listdir(folder_r)
        file_list = []
        for filename in os.listdir(folder_r):
            if filename[0:3] == 'res':
                file_list.append(filename)
        method_r_list = ['adapt', 'ihw']
        for method in method_r_list:
            summary_stats[method] = {}
            summary_stats[method]['FDP'] = np.zeros([n_data, len(alpha_list)])
            summary_stats[method]['Power'] = np.zeros([n_data, len(alpha_list)])
            for i_data,filename in enumerate(file_list):
                file_path = folder_r + '/' + filename
                temp_data = np.loadtxt(file_path, skiprows=1, delimiter = ',')
                h = temp_data[:, 0]
                for i_alpha,alpha in enumerate(alpha_list):                                                
                    if method == 'adapt':
                        h_hat = temp_data[:, i_alpha+1]
                    else:
                        h_hat = temp_data[:, i_alpha+5]
                    summary_stats[method]['FDP'][i_data, i_alpha] =\
                        np.sum((h==0)*(h_hat==1)) / max(np.sum(h_hat==1), 1)
                    summary_stats[method]['Power'][i_data, i_alpha] =\
                        np.sum((h==1)*(h_hat==1)) / np.sum(h==1)
    return summary_stats, time_dic

def plot_size_power(summary_stats, method_mapping_dic, data_name='', output_folder=None):
    marker_list = ['o', 'v', '^', '*', 'h', 'd']
    # color_list = ['C8', 'C5', 'C1', 'C2', 'C3', 'C0']
    color_list = ['C1', 'C2', 'C3', 'C0', 'C5', 'C8']
    alpha_list = [0.05, 0.1, 0.15, 0.2]
    axes = plt.figure(figsize = [5, 4])
    method_list = list(summary_stats.keys())
    method_list = ['nfdr (fast)', 'nfdr', 'adapt', 'ihw', 'sbh', 'bh']
    n_data = summary_stats[method_list[0]]['FDP'].shape[0]
    for i_method,method in enumerate(method_list):
        y_val = np.mean(summary_stats[method]['FDP'], axis=0)
        y_err = np.std(summary_stats[method]['FDP'], axis=0) / np.sqrt(n_data) * 1.96
        plt.errorbar(alpha_list, y_val, yerr=y_err, label=method_mapping_dic[method],\
                     capsize=4, elinewidth = 1.5, linewidth=1.5,\
                     color = color_list[i_method], marker = marker_list[i_method],\
                     markersize = 6, alpha=0.8)
    x_min, x_max = plt.xlim()
    y_min, y_max = plt.ylim()
    axis_min = min(x_min, y_min)
    axis_max = max(x_max, y_max)
    plt.plot([axis_min, axis_max], [axis_min, axis_max], linestyle='--', color='k')
    plt.legend(loc=2, fontsize=12)
    plt.ylabel('FDP', fontsize=16)
    plt.xlabel('nominal FDR', fontsize=16)
    if output_folder is not None:
        plt.tight_layout()
        plt.savefig(output_folder+'fdp_%s.png'%data_name)
        plt.savefig(output_folder+'fdp_%s.pdf'%data_name)
    else:
        plt.show()
    axes = plt.figure(figsize = [5, 4])
    for i_method,method in enumerate(method_list):
        y_val = np.mean(summary_stats[method]['Power'], axis=0)
        y_err = np.std(summary_stats[method]['Power'], axis=0) / np.sqrt(n_data) * 1.96
        plt.errorbar(alpha_list, y_val, yerr=y_err, label=method_mapping_dic[method],\
                     capsize=4, elinewidth = 1.5, linewidth=1.5,\
                     color = color_list[i_method], marker = marker_list[i_method],\
                     markersize = 6, alpha=0.8)
    plt.legend(loc=2, fontsize=12)
    plt.ylabel('power', fontsize=16)
    plt.xlabel('nominal FDR', fontsize=16)
    if output_folder is not None:
        plt.tight_layout()
        plt.savefig(output_folder+'power_%s.png'%data_name)
        plt.savefig(output_folder+'power_%s.pdf'%data_name)
    else:
        plt.show()
    plt.close('all')